/*
 * EMEngine.h
 *
 *  Created on: Jan 19, 2010
 *      Author: duyvu
 */

#ifndef EMENGINE_H_
#define EMENGINE_H_

#include "model/DegreeModel.h"

#include "util/StringTokenizer.h"

#include "iostream"
#include "fstream"

using namespace std;

namespace mixderg {

class EMEngine {
protected:
	DegreeModel* model;
	int modelType;
	int numOfEMRuns;
	int maxNumOfEMIterations;
	int maxNumOfESteps;

	double minTau;
	double minAlpha;

	double tauPrecision;
	double paramPrecision;

	int numOfClasses;

	// User needs to retrieve these values and set up the state of random generator.
	int seed1;
	int seed2;
public:
	EMEngine();
	virtual ~EMEngine();

	virtual void configure(const char* configurationFile, const char* delim);

	void setData(const char* dataFile, int fileFormat, const char* delim);

	void runEM();

	void runDetailedEM();

	void runInitializedEM(const char* tauFile, const char* betaFile);

	void runOneModel(string modelLLHFilename, string bestModelFilename,
			int outputPrecision);

	void runOneModel(string modelLLHFilename, string bestModelFilename,
			string currentBestLLHFilename, int outputPrecision, int fromRun);

	void runDetailedOneModel(string modelICLFilename, string bestModelFilename,
			string bestICLFilename, int outputPrecision, int fromRun);

	double getParamPrecision() const {
		return paramPrecision;
	}

	double getTauPrecision() const {
		return tauPrecision;
	}

	void setParamPrecision(double paramPrecision) {
		this->paramPrecision = paramPrecision;
	}

	void setTauPrecision(double tauPrecision) {
		this->tauPrecision = tauPrecision;
	}

	int getNumOfVertices() const {
		if (model != NULL)
			return model->getNumOfVertices();
		else
			return -1;
	}

	int getNumOfEdges() const {
		if (model != NULL)
			return model->getNumOfEdges();
		else
			return -1;
	}

	int getNumOfClasses() const {
		return numOfClasses;
	}

	void setNumOfClasses(int numOfClusters) {
		this->numOfClasses = numOfClusters;
		if (model != NULL)
			model->setNumOfClasses(numOfClusters);
	}

	int getSeed1() const {
		return seed1;
	}

	void setSeed1(int seed1) {
		this->seed1 = seed1;
	}

	int getSeed2() const {
		return seed2;
	}

	void setSeed2(int seed2) {
		this->seed2 = seed2;
	}

	int getNumOfEMRuns() const {
		return numOfEMRuns;
	}

	void setNumOfEMRuns(int numOfEMRuns) {
		this->numOfEMRuns = numOfEMRuns;
	}

	int getMaxNumOfEMIterations() const {
		return maxNumOfEMIterations;
	}

	void setMaxNumOfEMIterations(int maxNumOfEMIterations) {
		this->maxNumOfEMIterations = maxNumOfEMIterations;
	}

	int getMaxNumOfESteps() const {
		return maxNumOfESteps;
	}

	void setMaxNumOfESteps(int maxNumOfESteps) {
		this->maxNumOfESteps = maxNumOfESteps;
	}

	double getMinTau() const {
		return minTau;
	}

	void setMinTau(double minTau) {
		this->minTau = minTau;
	}

	double getMinAlpha() const {
		return minAlpha;
	}

	void setMinAlpha(double minAlpha) {
		this->minAlpha = minAlpha;
	}
};

}

#endif /* EMENGINE_H_ */
