/*
 * DegreeModel.h
 *
 *  Created on: Oct 9, 2011
 *      Author: duyvu
 */

#ifndef DEGREEMODEL_H_
#define DEGREEMODEL_H_

#include "DataTypes.h"

#include "iostream"
#include "fstream"

namespace mixderg {

class DegreeModel {
protected:
	double minAlpha;
	double minTau;
	double minTheta;

public:
	unsigned int numOfVertices;
	unsigned int numOfEdges;
	unsigned int numOfClasses;

	DegreeModel();
	virtual ~DegreeModel();

	virtual void setData(const char* dataFile, int fileFormat,
			const char* delim) = 0;

	virtual void setModel() = 0;
	virtual void setModel(int _numOfClasses, double _minTau, double _minAlpha,
			double _minTheta) = 0;

	virtual void resetParameters() = 0;
	virtual void resetParameters(const UnsignedIntVector& Z) = 0;
	virtual void resetParametersPartially(const UnsignedIntVector& Z,
			double hardPercentage) = 0;
	virtual void resetParameters(const DoubleMatrix& initTau) = 0;

	virtual void runFixedPointEstimationEStep(int emIter, int eIter) = 0;
	virtual bool isTauSignificantlyChanged(double _tauPrecision) = 0;
	virtual double getLargestTauChange() = 0;

	virtual void runModelEstimationMStep(int emIter) = 0;
	virtual bool areAlphaPiSignificantlyChanged(double _paramPrecision) = 0;

	virtual double getLLHLowerBound() = 0;
	virtual double getICL() = 0;

	virtual void copyParameters(DegreeModel* model) = 0;
	virtual void writeModel2File(ofstream & modelFile) = 0;
	virtual void writeParameters2File(ofstream & modelFile) = 0;

	/* Receive a matrix N x K and return a vector K */
	void sumDoubleMatrixOverRows(const DoubleMatrix& matrix,
			DoubleVector& vector);

	/* Normalize tau so that sum tau_ik over k is 1 and each membership
	 * value should be at leat minValue
	 */
	void normalizeTau(DoubleMatrix& tau, double minValue);

	void normalizeVector(DoubleVector& vector, double minValue);

	/* Solve QP with one equality constraint and bounded constraints */
	void solveQP(const DoubleMatrix& m, const DoubleMatrix& s,
			DoubleMatrix& tau, double precision);

	double getMinAlpha() const {
		return minAlpha;
	}

	void setMinAlpha(double minAlpha) {
		this->minAlpha = minAlpha;
	}

	void setMinTau(double minTau) {
		this->minTau = minTau;
	}

	double getMinTau() const {
		return minTau;
	}

	void setMinTheta(double minPi) {
		this->minTheta = minPi;
	}

	double getMinTheta() const {
		return minTheta;
	}

	int getNumOfVertices() const {
		return numOfVertices;
	}

	int getNumOfEdges() const {
		return numOfEdges;
	}

	int getNumOfClasses() const {
		return numOfClasses;
	}

	virtual void setNumOfClasses(int _numOfClasses) {
		numOfClasses = _numOfClasses;
	}
};

}

#endif /* DEGREEMODEL_H_ */
