/*
 * BootstrapNRMMExcessTrustModel.cpp
 *
 *  Created on: Dec 25, 2011
 *      Author: duyvu
 */

#include "BootstrapNRMMExcessTrustModel.h"

#define MATHLIB_STANDALONE
#include <Rmath.h>

namespace mixderg {

BootstrapNRMMExcessTrustModel::BootstrapNRMMExcessTrustModel() {
}

BootstrapNRMMExcessTrustModel::~BootstrapNRMMExcessTrustModel() {
}

void BootstrapNRMMExcessTrustModel::resetParameters() {

	cout << "BootstrapNRMMExcessTrustModel::resetParameters()" << endl;

	for (unsigned int i = 0; i < numOfVertices; i++)
		for (unsigned int k = 0; k < numOfClasses; k++)
			tau[i][k] = 0;

	// copy alpha to an pointer-based array
	double * selectedProbs = new double[numOfClasses];
	for (unsigned int k = 0; k < numOfClasses; k++)
		selectedProbs[k] = 1.0 / numOfClasses;

	// generate the latent class membership
	int * z_i = new int[numOfClasses];
	for (unsigned int i = 0; i < numOfVertices; i++) {
		rmultinom(1, selectedProbs, numOfClasses, z_i);
		int selectedClass = 0;
		while (z_i[selectedClass] == 0)
			selectedClass++;
		tau[i][selectedClass] = 1;
	}

	// clean up some pointer-based vectors and matrices
	if (selectedProbs != NULL)
		delete selectedProbs;
	if (z_i != NULL)
		delete z_i;

	normalizeTau(tau, minTau);

	cout << "Resetting model parameters" << endl;
	for (unsigned int k = 0; k < numOfClasses; k++) {
		alpha[k] = 0;
		excessTrustTheta[k] = 0;
	}
	negDenTheta = 0;
	negRecTheta = 0;
	posRecTheta = 0;
	updateKappa();

	runModelEstimationMStep(0);
}

}
