/*
 * MMExcessTrustModel.h
 *
 *  Created on: Dec 9, 2011
 *      Author: duyvu
 */

#ifndef MMEXCESSTRUSTMODEL_H_
#define MMEXCESSTRUSTMODEL_H_

#include "model/DegreeModel.h"

namespace mixderg {

class MMExcessTrustModel: public mixderg::DegreeModel {
protected:
	// Class memberships for nodes
	DoubleMatrix prevTau;

	// Class mixing probabilities;
	DoubleVector prevAlpha;

	// Theta paramters
	DoubleVector prev_excessTrustTheta;
	double prev_negDenTheta, prev_negRecTheta, prev_posRecTheta;

	void computeNetworkStatistics(int i, int j, int value,
			VectorOfNodeSet & negFollowers, VectorOfNodeSet & posFollowers);

	void updateKappa();
	void updateIndependentKappa();
	void updateGamma();
	void updateIndependentGamma();
	void updateGamma1();
	void updateGamma2();
	void updateGamma3();
	void updateGamma4Minus();
	void updateGamma4Plus();

public:
	// Class memberships for nodes
	DoubleMatrix tau;

	// Class mixing probabilities;
	DoubleVector alpha;

	// Theta paramters
	DoubleVector excessTrustTheta;
	double negDenTheta, negRecTheta, posRecTheta;

	// Gamma terms
	DoubleMatrix kappa, gamma, gamma1, gamma2, gamma3, gamma4Minus, gamma4Plus;

	// Network statistics
	DoubleVector excessTrustDegrees;
	double negDensity, negReciprocity, posReciprocity;

	MMExcessTrustModel();
	virtual ~MMExcessTrustModel();

	void setData(const char* dataFile, int fileFormat, const char* delim);

	void setModel();
	void setModel(int _numOfClasses, double _minTau, double _minAlpha,
			double _minTheta);

	void resetParameters();
	void resetParameters(const UnsignedIntVector& Z);
	void resetParametersPartially(const UnsignedIntVector& Z,
			double hardPercentage);
	void resetParameters(const DoubleMatrix& initTau);

	void runFixedPointEstimationEStep(int emIter, int eIter);
	bool isTauSignificantlyChanged(double _tauPrecision);
	double getLargestTauChange();

	virtual void runModelEstimationMStep(int emIter) = 0;
	bool areAlphaPiSignificantlyChanged(double _paramPrecision);

	double computeCarryOverTerm(const DoubleMatrix & delta);
	double computeCarryOverTermByClass(unsigned int classIndex,
			const DoubleMatrix & delta);
	double computeWeightedDegrees(unsigned int classIndex,
			const DoubleVector & degrees);

	double getLLHLowerBound();
	double getICL();

	void copyParameters(DegreeModel* model);
	void writeModel2File(ofstream & modelFile);
	void writeParameters2File(ofstream & modelFile);

};

}

#endif /* MMEXCESSTRUSTMODEL_H_ */
