/*
 * NRMMExcessTrustModel.h
 *
 *  Created on: Mar 9, 2012
 *      Author: duyvu
 */

#ifndef NRMMEXCESSTRUSTMODEL_H_
#define NRMMEXCESSTRUSTMODEL_H_

#include "MMExcessTrustModel.h"

#include <math.h>
#include "math/matrix.h"
#include "math/ludcmp.h"

namespace mixderg {

class NRMMExcessTrustModel: public mixderg::MMExcessTrustModel {
protected:
	int numStats;
	int numDyads;
	int numParams;
	Matrix statDyadTest;
	DoublePointerMatrix dyadProbs;
	Matrix A_t;

	void copyParams2Thetas(const DoubleVector & params);
	void copyThetas2Params(DoubleVector & params);
	void copyParams2Thetas(const DoubleVector & params,
			DoubleVector & excessTrustTheta, double & negDenTheta,
			double & negRecTheta, double & posRecTheta);

	void computeCachingParameters(const DoubleVector & excessTrustTheta,
			const double & negDenTheta, const double & negRecTheta,
			const double & posRecTheta, DoubleMatrix & gammaMinus0,
			DoubleMatrix & gammaPlus0, DoubleMatrix & gammaMinusPlus,
			DoubleMatrix & gammaRecMinus, DoubleMatrix & gammaRecPlus,
			DoubleMatrix & gamma, DoubleMatrix & kappa);

	void runNewtonRaphson(DoubleVector& params, int maxSteps,
			double relLLHTolerance, double scoreTolerance);
	double getLogLLH(const DoubleVector& params);

	void getScoreNegHessian(const DoubleVector& params, Vec& u, Matrix& I);
	void computeDyadProbs(const DoubleMatrix & gammaMinus0,
			const DoubleMatrix & gammaPlus0,
			const DoubleMatrix & gammaMinusPlus,
			const DoubleMatrix & gammaRecMinus,
			const DoubleMatrix & gammaRecPlus, const DoubleMatrix & gamma);
	double computeAlphaProduct(unsigned int k, unsigned int l);
	void computeStatMean(Vec & statMean, unsigned int k, unsigned int l);
	void computeStatCov(Matrix & statCov, unsigned int k, unsigned int l);
public:
	NRMMExcessTrustModel();
	virtual ~NRMMExcessTrustModel();

	void runModelEstimationMStep(int emIter);

	void setNumOfClasses(int _numOfClasses);
};

}

#endif /* NRMMEXCESSTRUSTMODEL_H_ */
