/*
 * ParametricBoostrapSampler.cpp
 *
 *  Created on: Oct 4, 2010
 *      Author: duyvu
 */

#include "ParametricBoostrapSampler.h"

#include "model/signed/MMExcessTrustModel.h"
#include "model/signed/BootstrapNRMMExcessTrustModel.h"

#include "util/StringTokenizer.h"

#include <iostream>
#include <fstream>
#include <string>
#include <cfloat>
#include <sstream>

using namespace std;

namespace mixderg {

ParametricBoostrapSampler::ParametricBoostrapSampler() {
	// TODO Auto-generated constructor stub
	model = NULL;
}

ParametricBoostrapSampler::~ParametricBoostrapSampler() {
	// TODO Auto-generated destructor stub
	if (model != NULL)
		delete model;
}

void ParametricBoostrapSampler::normalizeDyadProbs() {

}

void ParametricBoostrapSampler::generateSamples(unsigned int numOfSamples,
		const char* outputDirectory, unsigned outputPrecision) {

	cout << "ParametricBoostrapSampler::generateSamples" << endl;

	// alpha
	ofstream outputAlphaFile;
	string outputAlphaFilename = string(outputDirectory) + "/alpha.txt";
	outputAlphaFile.open(outputAlphaFilename.c_str());
	if (outputAlphaFile.is_open()) {
		outputAlphaFile.precision(outputPrecision);
		for (unsigned int k = 0; k < numOfClasses; k++) {
			outputAlphaFile << alpha[k] << "\t";
		}
		outputAlphaFile.close();
	} else
		cout << "Can not open the file " << outputAlphaFilename << endl;

	// Thetas
	string outputParamsFilename = string(outputDirectory) + "/Theta.txt";
	ofstream outputParamsFile;
	outputParamsFile.open(outputParamsFilename.c_str());
	if (outputParamsFile.is_open()) {
		outputParamsFile.precision(outputPrecision);
		printModel2File(outputParamsFile);
		outputParamsFile.close();
	} else
		cout << "Can not open the file " << outputParamsFilename << endl;

	EdgeList Y;
	UnsignedIntVector Z = UnsignedIntVector(numOfVertices);
	for (unsigned int sampleIndex = 0; sampleIndex < numOfSamples; sampleIndex++) {

		std::cout << "Generating the sample " << sampleIndex << std::endl;

		// get the sample
		generateSample(Y, Z);

		// Z
		ofstream outputZFile;
		stringstream sSampleIndex;
		sSampleIndex << sampleIndex;
		string outputZFilename = string(outputDirectory) + "/Z_"
				+ sSampleIndex.str() + ".txt";
		outputZFile.open(outputZFilename.c_str());
		if (outputZFile.is_open()) {
			outputZFile << numOfVertices << "\t" << numOfClasses << endl;
			for (unsigned int i = 0; i < numOfVertices; i++)
				outputZFile << Z[i] << endl;
			outputZFile.close();
		} else
			cout << "Can not open the file " << outputZFilename << endl;

		// Y
		ofstream outputYFile;
		string outputYFilename = string(outputDirectory) + "/Y_"
				+ sSampleIndex.str() + ".txt";
		outputYFile.open(outputYFilename.c_str());
		if (outputYFile.is_open()) {
			outputYFile << numOfVertices << endl;
			EdgeList::iterator it;
			for (it = Y.begin(); it != Y.end(); it++) {
				Edge edge = *it;
				outputYFile << edge.sender << "\t" << edge.receiver << "\t"
						<< edge.value << endl;
			}
			outputYFile.close();
		} else
			cout << "Can not open the file " << outputYFilename << endl;
	}
}

void ParametricBoostrapSampler::generateSamples(unsigned fromSampleIndex,
		unsigned toSampleIndex, const char* outputDirectory,
		unsigned outputPrecision) {

	// alpha
	ofstream outputAlphaFile;
	string outputAlphaFilename = string(outputDirectory) + "/alpha.txt";
	outputAlphaFile.open(outputAlphaFilename.c_str());
	if (outputAlphaFile.is_open()) {
		outputAlphaFile.precision(outputPrecision);
		for (unsigned int k = 0; k < numOfClasses; k++) {
			outputAlphaFile << alpha[k] << "\t";
		}
		outputAlphaFile.close();
	} else
		cout << "Can not open the file " << outputAlphaFilename << endl;

	// Thetas
	string outputParamsFilename = string(outputDirectory) + "/Theta.txt";
	ofstream outputParamsFile;
	outputParamsFile.open(outputParamsFilename.c_str());
	if (outputParamsFile.is_open()) {
		outputParamsFile.precision(outputPrecision);
		printModel2File(outputParamsFile);
		outputParamsFile.close();
	} else
		cout << "Can not open the file " << outputParamsFilename << endl;

	EdgeList Y;
	UnsignedIntVector Z = UnsignedIntVector(numOfVertices);
	for (unsigned int sampleIndex = fromSampleIndex; sampleIndex
			<= toSampleIndex; sampleIndex++) {

		std::cout << "Generating the sample " << sampleIndex << std::endl;

		// get the sample
		generateSample(Y, Z);

		// Z
		ofstream outputZFile;
		stringstream sSampleIndex;
		sSampleIndex << sampleIndex;
		string outputZFilename = string(outputDirectory) + "/Z_"
				+ sSampleIndex.str() + ".txt";
		outputZFile.open(outputZFilename.c_str());
		if (outputZFile.is_open()) {
			outputZFile << numOfVertices << "\t" << numOfClasses << endl;
			for (unsigned int i = 0; i < numOfVertices; i++)
				outputZFile << Z[i] << endl;
			outputZFile.close();
		} else
			cout << "Can not open the file " << outputZFilename << endl;

		// Y
		ofstream outputYFile;
		string outputYFilename = string(outputDirectory) + "/Y_"
				+ sSampleIndex.str() + ".txt";
		outputYFile.open(outputYFilename.c_str());
		if (outputYFile.is_open()) {
			outputYFile << numOfVertices << endl;
			EdgeList::iterator it;
			for (it = Y.begin(); it != Y.end(); it++) {
				Edge edge = *it;
				outputYFile << edge.sender << "\t" << edge.receiver << "\t"
						<< edge.value << endl;
			}
			outputYFile.close();
		} else
			cout << "Can not open the file " << outputYFilename << endl;
	}
}

void ParametricBoostrapSampler::configure(const char* configurationFilename,
		const char* delim) {

	std::cout << "The configuration file is " << configurationFilename
			<< std::endl;

	std::map<std::string, std::string> configurationMap;
	std::string line;
	std::ifstream fileStream(configurationFilename);
	if (fileStream.is_open()) {
		while (!fileStream.eof()) {
			getline(fileStream, line);
			StringTokenizer strtok = StringTokenizer(line, delim);
			if (strtok.countTokens() >= 2) {
				std::string key = strtok.nextToken();
				std::string value = strtok.nextToken();
				configurationMap[key] = value;
			}
		}
	} else
		std::cout << "can not read the configuration file" << std::endl;

	for (std::map<std::string, std::string>::const_iterator it =
			configurationMap.begin(); it != configurationMap.end(); ++it) {
		std::cout << "(" << it->first << " , " << it->second << ")"
				<< std::endl;
	}

	modelType = atoi(configurationMap["Model_Type"].c_str());
	switch (modelType) {
	case BOOTSTRAP_NR_MM_EXCESS_TRUST_GRAPH:
		model = new BootstrapNRMMExcessTrustModel();
		break;
	default:
		std::cout << "Model type must be specified!!!" << std::endl;
		exit(-1);
	}

	numOfEMRuns = 1;

	if (configurationMap.count("Max_Num_Of_E_Steps") == 1)
		maxNumOfESteps = atoi(configurationMap["Max_Num_Of_E_Steps"].c_str());
	else
		maxNumOfESteps = 1;

	if (configurationMap.count("Max_Num_Of_EM_Iterations") == 1)
		maxNumOfEMIterations = atoi(
				configurationMap["Max_Num_Of_EM_Iterations"].c_str());
	else
		maxNumOfEMIterations = 1000;

	if (configurationMap.count("Min_Tau") == 1)
		minTau = atof(configurationMap["Min_Tau"].c_str());
	else
		minTau = 1e-10;
	model->setMinTau(minTau);

	if (configurationMap.count("Min_Alpha") == 1)
		minAlpha = atof(configurationMap["Min_Alpha"].c_str());
	else
		minAlpha = 1e-6;
	model->setMinAlpha(minAlpha);

	if (configurationMap.count("Tau_Precision") == 1)
		tauPrecision = atof(configurationMap["Tau_Precision"].c_str());
	else
		tauPrecision = 1e-6;

	if (configurationMap.count("ParamPrecision") == 1)
		paramPrecision = atof(configurationMap["ParamPrecision"].c_str());
	else
		paramPrecision = 1e-6;

	if (configurationMap.count("Randomized_Init") == 1) {
		randomizedInit = atoi(configurationMap["Randomized_Init"].c_str());
	} else
		randomizedInit = 0; // 0 for hard, 1 partial, 2 random
}

void ParametricBoostrapSampler::runEM(const UnsignedIntVector& Z) {
	cout << "ParametricBoostrapSampler::runEM" << endl;

	// Assume that model has been created
	// i.e. data and all parameters have been configured

	// Print out all convergence information of each bootstrap sample estimation run
	cout.setf(ios::fixed, ios::floatfield);
	cout.precision(20);

	// Initialize membership parameters
	switch (randomizedInit) {
	case 0:
		model->resetParameters(Z);
		break;
	case 1:
		model->resetParametersPartially(Z, HARD_PERCENTAGE);
		break;
	case 2:
		model->resetParameters();
		break;
	default:
		cout << "Wrong INIT CODE for BOOTSTRAP" << endl;
		return;
	}

	double lastLLHLowerBound = model->getLLHLowerBound();
	cout << "Initial Lower Bound = " << lastLLHLowerBound << endl;

	cout << "Starting a EM run" << endl;
	MMExcessTrustModel* myModel = (MMExcessTrustModel*) model;
	for (int emIteration = 0; emIteration < maxNumOfEMIterations; emIteration++) {

		// E Step
		for (int eStep = 0; eStep < maxNumOfESteps; eStep++)
			model->runFixedPointEstimationEStep(emIteration, eStep);

		// M Step
		model->runModelEstimationMStep(emIteration);

		// Covergence Details
		double newLLHLowerBound = model->getLLHLowerBound();
		double largestTauChange = model->getLargestTauChange();
		cout << emIteration << "\t";
		cout << newLLHLowerBound << "\t";
		cout << largestTauChange << "\t";
		for (unsigned int k = 0; k < model->numOfClasses; k++)
			cout << myModel->excessTrustTheta[k] << "\t";
		cout << myModel->negDenTheta << "\t";
		cout << myModel->negRecTheta << "\t";
		cout << myModel->posRecTheta << endl;

		// Convergence Criterion Check
		if ((emIteration > 0) && (fabs(
				(long double) (newLLHLowerBound - lastLLHLowerBound)
						/ newLLHLowerBound) < 1e-8)) {
			cout << "Break" << endl;
			break;
		} else
			lastLLHLowerBound = newLLHLowerBound;
	}
}

void ParametricBoostrapSampler::runEMWithFixedIterations(
		const UnsignedIntVector& Z) {
	cout << "ParametricBoostrapSampler::runEMWithFixedIterations" << endl;

	// Assume that model has been created
	// i.e. data and all parameters have been configured

	// Print out all convergence information of each bootstrap sample estimation run
	cout.setf(ios::fixed, ios::floatfield);
	cout.precision(20);

	// Initialize membership parameters
	switch (randomizedInit) {
	case 0:
		model->resetParameters(Z);
		break;
	case 1:
		model->resetParametersPartially(Z, HARD_PERCENTAGE);
		break;
	case 2:
		model->resetParameters();
		break;
	default:
		cout << "Wrong INIT CODE for BOOTSTRAP" << endl;
		return;
	}

	double lastLLHLowerBound = model->getLLHLowerBound();
	cout << "Initial Lower Bound = " << lastLLHLowerBound << endl;

	cout << "Starting a EM run" << endl;
	MMExcessTrustModel* myModel = (MMExcessTrustModel*) model;
	for (int emIteration = 0; emIteration < maxNumOfEMIterations; emIteration++) {

		// E Step
		for (int eStep = 0; eStep < maxNumOfESteps; eStep++)
			model->runFixedPointEstimationEStep(emIteration, eStep);

		// M Step
		model->runModelEstimationMStep(emIteration);

		// Covergence Details
		double newLLHLowerBound = model->getLLHLowerBound();
		double largestTauChange = model->getLargestTauChange();
		cout << emIteration << "\t";
		cout << newLLHLowerBound << "\t";
		cout << largestTauChange << "\t";
		for (unsigned int k = 0; k < model->numOfClasses; k++)
			cout << myModel->excessTrustTheta[k] << "\t";
		cout << myModel->negDenTheta << "\t";
		cout << myModel->negRecTheta << "\t";
		cout << myModel->posRecTheta << endl;

		lastLLHLowerBound = newLLHLowerBound;
	}
}

void ParametricBoostrapSampler::getBestEMRun(const UnsignedIntVector& Z,
		DegreeModel* bestModel, double& bestCLL, double& bestICL) {
	std::cout << "ParametricBoostrapSampler::getBestEMRun" << std::endl;
	runEM(Z);
	bestCLL = model->getLLHLowerBound();
	bestICL = model->getICL();
	bestModel->copyParameters(model);
}

void ParametricBoostrapSampler::getBestEMRunWithFixedIterations(
		const UnsignedIntVector& Z, DegreeModel* bestModel, double& bestCLL,
		double& bestICL) {
	std::cout << "ParametricBoostrapSampler::getBestEMRunWithFixedIterations"
			<< std::endl;
	runEMWithFixedIterations(Z);
	bestCLL = model->getLLHLowerBound();
	bestICL = model->getICL();
	bestModel->copyParameters(model);
}

void ParametricBoostrapSampler::estimateSample(unsigned sampleIndex,
		const string& inputDirectory, const string& outputDirectory,
		unsigned int outputPrecision, bool isIterationFixed) {

	clock_t start = clock();

	// This object will contain parameters of the best model
	DegreeModel* bestModel = NULL;
	switch (modelType) {
	case BOOTSTRAP_NR_MM_EXCESS_TRUST_GRAPH:
		bestModel = new BootstrapNRMMExcessTrustModel();
		break;
	default:
		cout << "The MODEL TYPE need to be specified!!!" << endl;
		return;
	}
	stringstream sSampleIndex;
	sSampleIndex << sampleIndex;

	string inputYFilename = inputDirectory + "/Y_" + sSampleIndex.str()
			+ ".txt";
	model->setData(inputYFilename.c_str(), 0, "\t");

	model->setModel(numOfClasses, minTau, minAlpha, minPi);

	UnsignedIntVector Z = UnsignedIntVector(numOfVertices);
	ifstream inputZFile;
	string inputZFilename = inputDirectory + "/Z_" + sSampleIndex.str()
			+ ".txt";
	cout << "Z file is " << inputZFilename << endl;
	inputZFile.open(inputZFilename.c_str());
	if (inputZFile.is_open()) {
		string line;
		getline(inputZFile, line); // just ignore the numbers of vertices and classes
		unsigned int index = 0;
		while (!inputZFile.eof() && index < numOfVertices) {
			getline(inputZFile, line);
			stringstream ss(line);
			ss >> Z[index];
			index++;
		}
		if (index != numOfVertices)
			cout << "There are some missing membership values!!!" << endl;
		inputZFile.close();
	} else
		cout << "Can not open the file " << inputZFilename << endl;

	double bestCLL = 0;
	double bestICL = 0;
	if (isIterationFixed)
		getBestEMRunWithFixedIterations(Z, bestModel, bestCLL, bestICL);
	else
		getBestEMRun(Z, bestModel, bestCLL, bestICL);

	// IF WE NEED TO RESOLVE LABEL SWITCHING, DO IT HERE

	// print out parameters to the file
	ofstream outputParamsFile;
	string outputParamsFilename = outputDirectory + "/parameters"
			+ sSampleIndex.str() + ".txt";
	outputParamsFile.open(outputParamsFilename.c_str());
	if (outputParamsFile.is_open()) {
		outputParamsFile.precision(outputPrecision);
		outputParamsFile << bestCLL << "\t" << bestICL << endl;
		bestModel->writeParameters2File(outputParamsFile);
		outputParamsFile.close();
	} else
		cout << "Can not open the file " << outputParamsFilename << endl;

	clock_t jobTime = clock() - start;
	ofstream outputJobTimeFile;
	string outputJobTimeFilename = outputDirectory + "/JobTimes.txt";
	outputJobTimeFile.open(outputJobTimeFilename.c_str(), ios::app);
	if (outputJobTimeFile.is_open()) {
		time_t rawtime;
		time(&rawtime);
		jobTime = jobTime / (1e6); // to seconds
		unsigned int hours = jobTime / (60 * 60);
		jobTime = jobTime % (60 * 60);
		unsigned int minutes = jobTime / 60;
		unsigned int seconds = jobTime % 60;
		stringstream sJobTime;
		sJobTime << hours << ":";
		sJobTime << minutes << ":";
		sJobTime << seconds;
		// Sample index + Job time in hours:minutes:seconds s+ Done time
		outputJobTimeFile << sampleIndex << "\t";
		outputJobTimeFile << sJobTime.str() << "\t";
		outputJobTimeFile << ctime(&rawtime);
		outputJobTimeFile.close();
	} else
		cout << "Can not open the file " << outputJobTimeFilename << endl;
}

void ParametricBoostrapSampler::estimateSamples(unsigned fromSampleIndex,
		unsigned toSampleIndex, const string& inputDirectory,
		const string& outputDirectory, unsigned int outputPrecision,
		bool isIterationFixed) {
	for (unsigned sampleIndex = fromSampleIndex; sampleIndex <= toSampleIndex; sampleIndex++)
		estimateSample(sampleIndex, inputDirectory, outputDirectory,
				outputPrecision, isIterationFixed);
}

}
