/*
 * EfficientExcessTrustParametricBoostrapSampler.cpp
 *
 *  Created on: Dec 20, 2011
 *      Author: duyvu
 */

#include "EfficientExcessTrustParametricBoostrapSampler.h"

#define MATHLIB_STANDALONE
#include <Rmath.h>

namespace mixderg {

EfficientExcessTrustParametricBoostrapSampler::EfficientExcessTrustParametricBoostrapSampler() {
}

EfficientExcessTrustParametricBoostrapSampler::~EfficientExcessTrustParametricBoostrapSampler() {
}

void EfficientExcessTrustParametricBoostrapSampler::moreInitialization() {
	currentNonzeroGraph = EdgeSet(numOfVertices);
	for (unsigned int i = 0; i < numOfVertices; i++)
		currentNonzeroGraph[i] = NodeSet();
}

void EfficientExcessTrustParametricBoostrapSampler::normalizeDyadProbs() {
	for (unsigned int k = 0; k < numOfClasses; k++)
		for (unsigned int l = 0; l < numOfClasses; l++) {
			double sum = 0;
			for (unsigned int t = 0; t < 8; t++)
				sum += dyadProbs[k][l][t];
			for (unsigned int t = 0; t < 8; t++)
				dyadProbs[k][l][t] /= sum;
		}
}

int EfficientExcessTrustParametricBoostrapSampler::generateDyadicValue(
		unsigned int k, unsigned int l, int *& dyad_ij, EdgeList & Y,
		unsigned int i, unsigned int j) {

	rmultinom(1, dyadProbs[k][l], 8, dyad_ij);
	int selectedDyadValue = 0;
	while (dyad_ij[selectedDyadValue] == 0)
		selectedDyadValue++;

	Edge edge;
	// y_ij = [-1,0,1,0,-1,1,-1,1,0];
	// y_ji = [0,-1,0,1,1,-1,-1,1,0];
	switch (selectedDyadValue) {
	// do not set 0 value because the matrix is in the sparse format
	// if we set value 0, it causes troubles when we iterate through nonzero elements
	// even being setting a value 0, an element turns to be nonzero ;-(
	case 0:
		edge.sender = i;
		edge.receiver = j;
		edge.value = -1;
		Y.push_back(edge);
		currentNonzeroGraph[i].insert(j);
		break;
	case 1:
		edge.sender = j;
		edge.receiver = i;
		edge.value = -1;
		Y.push_back(edge);
		currentNonzeroGraph[j].insert(i);
		break;
	case 2:
		edge.sender = i;
		edge.receiver = j;
		edge.value = 1;
		Y.push_back(edge);
		currentNonzeroGraph[i].insert(j);
		break;
	case 3:
		edge.sender = j;
		edge.receiver = i;
		edge.value = 1;
		Y.push_back(edge);
		currentNonzeroGraph[j].insert(i);
		break;
	case 4:
		edge.sender = i;
		edge.receiver = j;
		edge.value = -1;
		Y.push_back(edge);
		currentNonzeroGraph[i].insert(j);
		edge.sender = j;
		edge.receiver = i;
		edge.value = 1;
		Y.push_back(edge);
		currentNonzeroGraph[j].insert(i);
		break;
	case 5:
		edge.sender = i;
		edge.receiver = j;
		edge.value = 1;
		Y.push_back(edge);
		currentNonzeroGraph[i].insert(j);
		edge.sender = j;
		edge.receiver = i;
		edge.value = -1;
		Y.push_back(edge);
		currentNonzeroGraph[j].insert(i);
		break;
	case 6:
		edge.sender = i;
		edge.receiver = j;
		edge.value = -1;
		Y.push_back(edge);
		currentNonzeroGraph[i].insert(j);
		edge.sender = j;
		edge.receiver = i;
		edge.value = -1;
		Y.push_back(edge);
		currentNonzeroGraph[j].insert(i);
		break;
	case 7:
		edge.sender = i;
		edge.receiver = j;
		edge.value = 1;
		Y.push_back(edge);
		currentNonzeroGraph[i].insert(j);
		edge.sender = j;
		edge.receiver = i;
		edge.value = 1;
		Y.push_back(edge);
		currentNonzeroGraph[j].insert(i);
		break;
	case 8:
		break;
	default:
		cout << "ERROR!!! Undefined dyad value!" << endl;
		break;
	}

	return selectedDyadValue;
}

void EfficientExcessTrustParametricBoostrapSampler::generateSelfBlock(
		unsigned int k, EdgeList& Y, unsigned int kStartIndex,
		unsigned int kN) {

	int * dyad_ij = new int[8];

	// draw the number of non-zero dyads
	long long size = ((long long) kN) * (kN - 1) / 2;
	unsigned int M = rbinom(size, 1 - dyadProbs[k][k][8]);

	// randomly select each dyad then generate its value
	for (unsigned int edgeIndex = 0; edgeIndex < M; edgeIndex++) {
		unsigned int i;
		unsigned int j;

		// randomly select dyad ij
		do {
			i = kStartIndex + floor(runif(0, 1) * kN);
			j = kStartIndex + floor(runif(0, 1) * kN);
		} while (i == j
				|| (currentNonzeroGraph[i].find(j)
						!= currentNonzeroGraph[i].end())
				|| (currentNonzeroGraph[j].find(i)
						!= currentNonzeroGraph[j].end()));

		// draw value for the selected dyad ij
		generateDyadicValue(k, k, dyad_ij, Y, i, j);
	}

	if (dyad_ij != NULL)
		delete dyad_ij;
}

void EfficientExcessTrustParametricBoostrapSampler::generateCrossBlock(
		unsigned int k, unsigned int l, EdgeList & Y, unsigned int kStartIndex,
		unsigned int kN, unsigned int lStartIndex, unsigned int lN) {

	int *dyad_ij = new int[8];

	// draw the number of non-zero dyads
	long long size = ((long long) kN) * lN;
	unsigned int M = rbinom(size, 1 - dyadProbs[k][l][8]);

	// randomly select each dyad then generate its value
	for (unsigned int edgeIndex = 0; edgeIndex < M; edgeIndex++) {
		unsigned int i;
		unsigned int j;

		// randomly select dyad ij
		do {
			i = kStartIndex + floor(runif(0, 1) * kN);
			j = lStartIndex + floor(runif(0, 1) * lN);
		} while ((currentNonzeroGraph[i].find(j) != currentNonzeroGraph[i].end())
				|| (currentNonzeroGraph[j].find(i)
						!= currentNonzeroGraph[j].end()));

		// draw value for the selected dyad ij
		generateDyadicValue(k, l, dyad_ij, Y, i, j);
	}

	if (dyad_ij != NULL)
		delete dyad_ij;
}

void EfficientExcessTrustParametricBoostrapSampler::generateSample(EdgeList& Y,
		UnsignedIntVector& Z) {

	// copy alpha to an pointer-based array
	double * mixingProbs = new double[numOfClasses];
	for (unsigned int k = 0; k < numOfClasses; k++)
		mixingProbs[k] = alpha[k];

	// generate the latent class membership
	Z.clear();
	int * z_i = new int[numOfClasses];
	rmultinom(numOfVertices, mixingProbs, numOfClasses, z_i);
	UnsignedIntVector startIndices = UnsignedIntVector(numOfClasses);
	int startIndex = 0;
	for (unsigned int k = 0; k < numOfClasses; k++) {
		startIndices[k] = startIndex;
		unsigned int endIndex = startIndex + z_i[k];
		for (unsigned int i = startIndex; i < endIndex; i++)
			Z[i] = k;
		startIndex = endIndex;
	}

	// generate the edge list
	Y.clear();
	for (unsigned int i = 0; i < numOfVertices; i++)
		currentNonzeroGraph[i].clear();
	for (unsigned int k = 0; k < numOfClasses; k++) {
		for (unsigned int l = 0; l < k; l++) {
			generateCrossBlock(k, l, Y, startIndices[k], z_i[k],
					startIndices[l], z_i[l]);
		}
		generateSelfBlock(k, Y, startIndices[k], z_i[k]);
	}

	// clean up some pointer-based vectors and matrices
	if (mixingProbs != NULL)
		delete mixingProbs;
	if (z_i != NULL)
		delete z_i;
}

}
