/*
 * ExcessTrustParametricBoostrapSampler.cpp
 *
 *  Created on: Dec 20, 2011
 *      Author: duyvu
 */

#include "model/signed/MMExcessTrustModel.h"

#include "ExcessTrustParametricBoostrapSampler.h"

#include "util/StringTokenizer.h"

#include <iostream>
#include <fstream>
#include <string>

#define MATHLIB_STANDALONE
#include <Rmath.h>

using namespace std;

namespace mixderg {

ExcessTrustParametricBoostrapSampler::ExcessTrustParametricBoostrapSampler() {
}

ExcessTrustParametricBoostrapSampler::~ExcessTrustParametricBoostrapSampler() {
	for (unsigned int k = 0; k < numOfClasses; k++)
		for (unsigned int l = 0; l < numOfClasses; l++)
			if (dyadProbs[k][l] != NULL)
				delete dyadProbs[k][l];
}

void ExcessTrustParametricBoostrapSampler::setParameters(
		const char* modelFilename, const char* delim) {

	ifstream modelFile;
	modelFile.open(modelFilename);
	if (modelFile.is_open()) {
		string line;
		cout << "Reading the model file " << modelFilename << endl;

		getline(modelFile, line);
		cout << "First line " << line << endl;
		StringTokenizer tokenizer = StringTokenizer(line, delim);
		numOfVertices = tokenizer.nextIntToken();
		cout << "numOfVertices = " << numOfVertices << endl;

		getline(modelFile, line);
		cout << "Second line " << line << endl;
		tokenizer = StringTokenizer(line, delim);
		numOfClasses = tokenizer.nextIntToken();
		cout << "numOfClasses = " << numOfClasses << endl;

		getline(modelFile, line);
		cout << "Third line " << line << endl;
		tokenizer = StringTokenizer(line, delim);
		unsigned colIndex = 0;
		alpha = DoubleVector(numOfClasses);
		while (tokenizer.hasMoreTokens()) {
			alpha[colIndex++] = tokenizer.nextDoubleToken();
			cout << alpha[colIndex - 1] << "\t";
		}
		cout << endl;

		getline(modelFile, line);
		cout << "Forth line " << line << endl;
		tokenizer = StringTokenizer(line, delim);
		colIndex = 0;
		excessTrustTheta = DoubleVector(numOfClasses);
		while (tokenizer.hasMoreTokens()) {
			excessTrustTheta[colIndex++] = tokenizer.nextDoubleToken();
		}

		getline(modelFile, line);
		cout << "Fifth line " << line << endl;
		tokenizer = StringTokenizer(line, delim);
		negDenTheta = tokenizer.nextDoubleToken();

		getline(modelFile, line);
		cout << "Sixth line " << line << endl;
		tokenizer = StringTokenizer(line, delim);
		negRecTheta = tokenizer.nextDoubleToken();

		getline(modelFile, line);
		cout << "Seventh line " << line << endl;
		tokenizer = StringTokenizer(line, delim);
		posRecTheta = tokenizer.nextDoubleToken();

		DoubleMatrix gammaMinus0 = DoubleMatrix(numOfClasses,
				DoubleVector(numOfClasses));
		DoubleMatrix gammaPlus0 = DoubleMatrix(numOfClasses,
				DoubleVector(numOfClasses));
		DoubleMatrix gammaMinusPlus = DoubleMatrix(numOfClasses,
				DoubleVector(numOfClasses));
		DoubleMatrix gammaRecMinus = DoubleMatrix(numOfClasses,
				DoubleVector(numOfClasses));
		DoubleMatrix gammaRecPlus = DoubleMatrix(numOfClasses,
				DoubleVector(numOfClasses));
		DoubleMatrix gamma = DoubleMatrix(numOfClasses,
				DoubleVector(numOfClasses));
		for (unsigned int k = 0; k < numOfClasses; k++)
			for (unsigned int l = 0; l <= k; l++) {

				gammaMinus0[k][l] = exp(negDenTheta - excessTrustTheta[l]);
				gammaMinus0[l][k] = exp(negDenTheta - excessTrustTheta[k]);

				gammaPlus0[k][l] = exp(excessTrustTheta[l]);
				gammaPlus0[l][k] = exp(excessTrustTheta[k]);

				gammaMinusPlus[k][l]
						= exp(
								negDenTheta + excessTrustTheta[k]
										- excessTrustTheta[l]);
				gammaMinusPlus[l][k]
						= exp(
								negDenTheta + excessTrustTheta[l]
										- excessTrustTheta[k]);

				gammaRecMinus[k][l] = exp(
						2 * negDenTheta + negRecTheta - excessTrustTheta[k]
								- excessTrustTheta[l]);
				gammaRecMinus[l][k] = gammaRecMinus[k][l]; // symmetric

				gammaRecPlus[k][l]
						= exp(
								posRecTheta + excessTrustTheta[k]
										+ excessTrustTheta[l]);
				gammaRecPlus[l][k] = gammaRecPlus[k][l]; // symmetric

				gamma[k][l] = 1 + gammaMinus0[k][l] + gammaMinus0[l][k]
						+ gammaPlus0[k][l] + gammaPlus0[l][k]
						+ gammaMinusPlus[k][l] + gammaMinusPlus[l][k]
						+ gammaRecMinus[k][l] + gammaRecPlus[k][l];
				gamma[l][k] = gamma[k][l]; // symmetric
			}

		dyadProbs = DoublePointerMatrix(numOfClasses,
				DoublePointerVector(numOfClasses));
		for (unsigned int k = 0; k < numOfClasses; k++)
			for (unsigned int l = 0; l < numOfClasses; l++) {
				dyadProbs[k][l] = new double[9];
				// y_ij = [-1,0,1,0,-1,1,-1,1,0];
				// y_ji = [0,-1,0,1,1,-1,-1,1,0];
				// p0(k,l),p0(l,k),p1(k,l),p1(l,k),p2(k,l),p2(l,k),p3(k,l),p4(k,l),p5(k,l)
				dyadProbs[k][l][0] = gammaMinus0[k][l] / gamma[k][l];
				dyadProbs[k][l][1] = gammaMinus0[l][k] / gamma[k][l];
				dyadProbs[k][l][2] = gammaPlus0[k][l] / gamma[k][l];
				dyadProbs[k][l][3] = gammaPlus0[l][k] / gamma[k][l];
				dyadProbs[k][l][4] = gammaMinusPlus[k][l] / gamma[k][l];
				dyadProbs[k][l][5] = gammaMinusPlus[l][k] / gamma[k][l];
				dyadProbs[k][l][6] = gammaRecMinus[k][l] / gamma[k][l];
				dyadProbs[k][l][7] = gammaRecPlus[k][l] / gamma[k][l];
				dyadProbs[k][l][8] = 1.0 / gamma[k][l];
			}
		normalizeDyadProbs();

		modelFile.close();
	} else
		cout << "Can not open the file " << modelFilename << endl;

	moreInitialization();
}

void ExcessTrustParametricBoostrapSampler::moreInitialization() {
	return;
}

void ExcessTrustParametricBoostrapSampler::generateSample(EdgeList& Y,
		UnsignedIntVector& Z) {

	// copy alpha to an pointer-based array
	double * mixingProbs = new double[numOfClasses];
	for (unsigned int k = 0; k < numOfClasses; k++)
		mixingProbs[k] = alpha[k];

	// generate the latent class membership
	Z.clear();
	int * z_i = new int[numOfClasses];
	for (unsigned int i = 0; i < numOfVertices; i++) {
		rmultinom(1, mixingProbs, numOfClasses, z_i);
		int selectedClass = 0;
		while (z_i[selectedClass] == 0)
			selectedClass++;
		Z[i] = selectedClass;
	}

	// generate the edge list
	Y.clear();
	Edge edge;
	int * dyad_ij = new int[9];
	for (unsigned int i = 0; i < numOfVertices; i++)
		for (unsigned int j = 0; j < i; j++) {
			rmultinom(1, dyadProbs[Z[i]][Z[j]], 9, dyad_ij);
			int selectedDyadValue = 0;
			while (dyad_ij[selectedDyadValue] == 0)
				selectedDyadValue++;
			// y_ij = [-1,0,1,0,-1,1,-1,1,0];
			// y_ji = [0,-1,0,1,1,-1,-1,1,0];
			switch (selectedDyadValue) {
			// do not set 0 value because the matrix is in the sparse format
			// if we set value 0, it causes troubles when we iterate through nonzero elements
			// even being setting a value 0, an element turns to be nonzero ;-(
			case 0:
				edge.sender = i;
				edge.receiver = j;
				edge.value = -1;
				Y.push_back(edge);
				break;
			case 1:
				edge.sender = j;
				edge.receiver = i;
				edge.value = -1;
				Y.push_back(edge);
				break;
			case 2:
				edge.sender = i;
				edge.receiver = j;
				edge.value = 1;
				Y.push_back(edge);
				break;
			case 3:
				edge.sender = j;
				edge.receiver = i;
				edge.value = 1;
				Y.push_back(edge);
				break;
			case 4:
				edge.sender = i;
				edge.receiver = j;
				edge.value = -1;
				Y.push_back(edge);
				edge.sender = j;
				edge.receiver = i;
				edge.value = 1;
				Y.push_back(edge);
				break;
			case 5:
				edge.sender = i;
				edge.receiver = j;
				edge.value = 1;
				Y.push_back(edge);
				edge.sender = j;
				edge.receiver = i;
				edge.value = -1;
				Y.push_back(edge);
				break;
			case 6:
				edge.sender = i;
				edge.receiver = j;
				edge.value = -1;
				Y.push_back(edge);
				edge.sender = j;
				edge.receiver = i;
				edge.value = -1;
				Y.push_back(edge);
				break;
			case 7:
				edge.sender = i;
				edge.receiver = j;
				edge.value = 1;
				Y.push_back(edge);
				edge.sender = j;
				edge.receiver = i;
				edge.value = 1;
				Y.push_back(edge);
				break;
			case 8:
				break;
			default:
				cout << "ERROR!!! Undefined dyad value!" << endl;
				break;
			}
		}

	// clean up some pointer-based vectors and matrices
	if (mixingProbs != NULL)
		delete mixingProbs;
	if (z_i != NULL)
		delete z_i;
	if (dyad_ij != NULL)
		delete dyad_ij;
}

void ExcessTrustParametricBoostrapSampler::printModel2File(ofstream & modelFile) {
	// Class mixing probabilities;
	for (unsigned k = 0; k < numOfClasses; k++)
		modelFile << alpha[k] << "\t";
	modelFile << endl;
	// Degree paramters
	for (unsigned k = 0; k < numOfClasses; k++)
		modelFile << excessTrustTheta[k] << "\t";
	modelFile << endl;
	modelFile << negDenTheta << endl;
	modelFile << negRecTheta << endl;
	modelFile << posRecTheta << endl;
}

void ExcessTrustParametricBoostrapSampler::initializeModelParameters() {
	MMExcessTrustModel* myModel = (MMExcessTrustModel*) model;
	for (unsigned int k = 0; k < numOfClasses; k++)
		myModel->excessTrustTheta[k] = excessTrustTheta[k];
	myModel->negDenTheta = negDenTheta;
	myModel->negRecTheta = negRecTheta;
	myModel->posRecTheta = posRecTheta;
}

}
